import os
import sys
import json
import numpy as np
from typing import List

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from gext.fitting import LeastSquare, QuasiTimeReversible

def load_json(path):
    with open(path, "r") as json_file:
        data = json.load(json_file)

    for key, value in data.items():
        if isinstance(value, list):
            if isinstance(value[0], float):
                data[key] = np.array(value, dtype=np.float64)
            else:
                data[key] = np.array(value)
    return data

class AlternativeLeastSquare(LeastSquare):

    """Alternative least square minimization fitting."""

    def fit(self, vectors: List[np.ndarray], target: np.ndarray):
        """Given a set of vectors and a target return the fitting
        coefficients."""
        matrix = np.array(vectors).T
        a = np.vstack((matrix, np.identity(len(vectors))*self.options["regularization"]))
        b = np.concatenate((target, np.zeros(len(vectors))))
        coefficients, _, _, _ = np.linalg.lstsq(a, b, rcond=-1)
        return np.array(coefficients, dtype=np.float64)

class AlternativeQuasiTimeReversible(QuasiTimeReversible):

    """Quasi time reversible fitting scheme."""

    def fit(self, vectors: List[np.ndarray], target: np.ndarray):
        """Given a set of vectors and a target return the fitting
        coefficients in a quasi time reversible scheme."""

        past_target = vectors[0]
        matrix = np.array(vectors[1:]).T

        q = matrix.shape[1]
        if q == 1:
            time_reversible_matrix = matrix
        elif q%2 == 0:
            time_reversible_matrix = matrix[:, :q//2] + matrix[:, :q//2-1:-1]
        else:
            time_reversible_matrix = matrix[:, :q//2+1] + matrix[:, :q//2-1:-1]

        a = np.vstack((time_reversible_matrix,
            np.identity(time_reversible_matrix.shape[1])*self.options["regularization"]))
        b = np.concatenate((target + past_target, np.zeros((time_reversible_matrix.shape[1]))))

        coefficients, _, _, _ = np.linalg.lstsq(a, b, rcond=-1)


        if q == 1:
            full_coefficients = np.concatenate(([-1.0], coefficients))
        elif q%2 == 0:
            full_coefficients = np.concatenate(([-1.0], coefficients,
                coefficients[::-1]))
        else:
            full_coefficients = np.concatenate(([-1.0], coefficients[:-1],
                2.0*coefficients[-1:], coefficients[-2::-1]))
        return np.array(full_coefficients, dtype=np.float64)
